import re
from fractions import Fraction
from typing import Any, ClassVar

from ..exceptions import StackTypeError, StackValueError
from ..types import Boolean, Float, Integer, String, TokenAmount, User, Wrapper
from .stack_op import StackOp


class Swap(StackOp):
    def operation(
        self, arg1: Wrapper[Any], arg2: Wrapper[Any]
    ) -> tuple[Wrapper[Any], Wrapper[Any]]:
        return arg2, arg1

    def type_check(self, *args: Any) -> None:
        pass


class Dup(StackOp):
    def operation(self, arg: Wrapper[Any]) -> tuple[Wrapper[Any], Wrapper[Any]]:
        return arg, arg

    def type_check(self, *args: Any) -> None:
        pass


class GreaterThan(StackOp):
    def operation(self, arg1: Integer, arg2: Integer) -> Boolean:
        return Boolean(arg1.value > arg2.value)


class GreaterThanOrEqual(StackOp):
    def operation(self, arg1: Integer, arg2: Integer) -> Boolean:
        return Boolean(arg1.value >= arg2.value)


class Equal(StackOp):
    def operation(self, arg1: Integer, arg2: Integer) -> Boolean:
        return Boolean(arg1.value == arg2.value)


class Not(StackOp):
    def operation(self, arg: Boolean) -> Boolean:
        return Boolean(not arg.value)


class Push(StackOp):
    immediate_instruction: ClassVar[bool] = True

    def operation(self, arg: str) -> Wrapper[Any]:
        if arg.startswith('"') and arg.endswith('"'):
            return String(arg[1:-1])
        elif arg.startswith("@"):
            return User(arg)
        elif "." in arg and arg.replace(".", "").isdigit():
            return Float(float(arg))
        elif arg.isdigit():
            return Integer(int(arg))
        elif arg in ["true", "false"]:
            return Boolean(arg == "true")
        elif self.is_fraction_string(arg):
            return Float(float(Fraction(arg)))
        else:
            raise ValueError(f"Invalid value: {arg}")

    def is_fraction_string(self, value: str) -> bool:
        return bool(re.fullmatch(r"-?\d+/\d+", value))


class Pop(StackOp):
    def operation(self, arg: Wrapper[Any]) -> None:
        return None

    def type_check(self, *args: Any) -> None:
        pass


class Add(StackOp):
    def operation(
        self,
        a: Integer | Float | TokenAmount,
        b: Integer | Float | TokenAmount,
    ) -> Integer | Float | TokenAmount:
        if a.title == "token_amount" and b.title == "token_amount":
            if a.value[1].value == b.value[1].value:  # type: ignore[index]
                a.value[0].value += b.value[0].value  # type: ignore[index]
                return a
            else:
                raise StackValueError(
                    "Cannot add token amounts from different addresses."
                    f"{a.value[1].value} != {b.value[1].value}"  # type: ignore[index]
                )
        elif a.title == "token_amount":
            a.value[0].value += b.value  # type: ignore
        else:
            a.value += b.value  # type: ignore
        return a


class Sub(StackOp):
    def operation(
        self,
        a: Integer | Float | TokenAmount,
        b: Integer | Float | TokenAmount,
    ) -> Integer | Float | TokenAmount:
        if a.title == "token_amount":
            if isinstance(b, TokenAmount):  # type: ignore[misc]
                if not a.value[1].value == b.value[1].value:  # type: ignore[index]
                    raise StackValueError(
                        "Cannot subtract token amounts from different addresses"
                    )
                a.value[0].value -= b.value[0].value  # type: ignore
            else:
                a.value[0].value -= b.value  # type: ignore
        else:
            a.value -= b.value  # type: ignore
        return a


class Mul(StackOp):
    def operation(
        self,
        a: Integer | Float | TokenAmount,
        b: Integer | Float | TokenAmount,
    ) -> Integer | Float | TokenAmount:
        if a.title == "token_amount" and b.title == "token_amount":
            if a.value[1].value == b.value[1].value:  # type: ignore[index]
                a.value[0].value *= b.value[0].value  # type: ignore[index]
                return a
            else:
                raise StackValueError(
                    "Cannot multiply token amounts from different addresses"
                )
        elif a.title == "token_amount":
            a.value[0].value *= b.value  # type: ignore
        else:
            a.value *= b.value  # type: ignore
        return a


class Div(StackOp):
    def operation(
        self,
        a: Integer | Float | TokenAmount,
        b: Integer | Float | TokenAmount,
    ) -> Integer | Float | TokenAmount:
        if b.title == "token_amount":
            raise StackValueError(
                "The second value on the stack can not be a token amount"
            )
        elif a.title == "token_amount":
            a.value[0].value /= b.value  # type: ignore

        return Float(a.value / b.value)  # type: ignore[operator]


class Mod(StackOp):
    def operation(
        self,
        a: Integer,
        b: Integer,
    ) -> Integer:
        a.value %= b.value  # type: ignore
        return a


class Branch(StackOp):
    def operation(
        self, condition: Boolean, true_value: Wrapper[Any], false_value: Wrapper[Any]
    ) -> Wrapper[Any]:
        if condition.value:
            return true_value
        else:
            return false_value

    def type_check(self, *args: Any):
        if not isinstance(args[0], Boolean):  # type: ignore[misc]
            raise StackTypeError("The first value on the stack must be a boolean")


BASE_OPERATORS: list[type[StackOp]] = [
    Swap,
    Push,
    Pop,
    Add,
    Sub,
    Mul,
    Div,
    Mod,
    # Dup,
    Branch,
    GreaterThan,
    GreaterThanOrEqual,
    Equal,
    Not,
]



from .impls.executors import NoOpExecutor
from .interfaces import (
    AgentExecutor,
    Classifier,
    ClassifierResponse,
    Effect,
    Executor,
    PreProcess,
    Processor,
    SourceT,
)
from .logger import logger
from .pipeline import Pipeline

__all__ = [
    "AgentExecutor",
    "Classifier",
    "ClassifierResponse",
    "Effect",
    "Executor",
    "NoOpExecutor",
    "Pipeline",
    "PreProcess",
    "Processor",
    "SourceT",
    "logger",
]



from .base import BASE_OPERATORS
from .special import SPECIAL_OPERATORS
from .stack_op import StackOp

OPERATORS = BASE_OPERATORS + SPECIAL_OPERATORS


__all__ = [
    "BASE_OPERATORS",
    "OPERATORS",
    "SPECIAL_OPERATORS",
    "StackOp",
]



from typing import Annotated

from pydantic import BaseModel
from typing_extensions import Doc


class TransferArgs(BaseModel):
    token_address: Annotated[str, Doc("The address of the token to transfer")]
    amount: Annotated[int, Doc("The amount of the token to transfer")]

    recipient: Annotated[
        str, Doc("The twitter handle or address of the recipient of the transfer")
    ]



from textwrap import dedent
from typing import Any, Callable, Sequence

from pydantic._internal._model_construction import ModelMetaclass

from .example import EXAMPLE_PROGRAM
from .operators import SPECIAL_OPERATORS, StackOp
from .types import DEFAULT_TYPES, Wrapper


def generate_types_doc(types: Sequence) -> str:
    return "\n".join(
        [f"- `{t._annotation().title}`: {t._annotation().description}" for t in types]
    )


def convert_function_to_string(func: Callable[..., type[Wrapper[Any]]]) -> str:
    annotations = func.__annotations__
    func_name = func.__name__.upper()
    docstring = func.__doc__

    func_signature = ":: "
    for name, annotation in annotations.items():
        if name == "return":
            break

        arg_type = annotation._annotation().title
        func_signature += f"'{arg_type} : "
    func_signature += "'S -> "
    return_type = annotations["return"]

    if return_type is None:
        func_signature += "'S"
    else:
        func_signature += f"'{return_type._annotation().title} : 'S"
    return (
        dedent(
            f"""
        {func_name}: {docstring}
        {func_signature}
    """
        ).strip()
        + "\n"
    )


def make_syntax_doc(
    types: Sequence[ModelMetaclass] = DEFAULT_TYPES,
    examples: str = EXAMPLE_PROGRAM,
    special_operators: Sequence[type[StackOp]] = SPECIAL_OPERATORS,
) -> str:
    types_doc = generate_types_doc(types)
    special_operators_doc = "\n".join(op.to_docstring() for op in special_operators)
    return f"""
You are an agent that codes in "Dowse," a type-safe stack-based programming language featuring
a defined set of types and operators.  Dowse is unique from other stack-based languages, as you
can also assign external variables that can store them and they can be pushed onto the stack later.

In Dowse, elements are pushed onto the stack to the left, pushing existing elements to the right
side of the stack. This consistent ordering maintains the integrity of stack operations.

Each element on the stack is prefixed with a single quote to denote its presence. For example,
executing `PUSH "!"` updates the stack as follows:
// ['!:string]

Subsequently, executing `PUSH "?"` modifies the stack to:
// ['?:string '!:string]


** Assigning Variables **
You can assign a variable by executing the command:

ASSIGN variable_name

This will assign the top variable on the stack to the variable name.
This type can then be pushed onto the stack via:

PUSH &variable_name
// ['variable_name:variable_type]

For example:

PUSH cbBTC_address
// ['cbBTC_address:address]

ASSIGN tmp
// ['cbBTC_address:address]

PUSH &tmp
// ['cbBTC_address:address 'cbBTC_address:address]

This relaxes the formality of the stack, facilitating easier coding.

**Pushing Types:**

You can push types directly on to the stack if the type has as specific format.  For example,

PUSH 10
// ['10:integer]

PUSH 0.001
// ['0.001:float]

PUSH @user1
// ['@user1:user]

PUSH true
// ['true:boolean]

But you can not push an expression like:

PUSH 10 + 20
// ERROR: Invalid value: 10 + 20

**Comments:**

Comments start with "//" and go to the end of the line.

**Types:**

{types_doc}

**Basic Operators:**

SWAP:
:: 'a : 'b : 'S -> 'b : 'a : 'S

PUSH 'a
:: 'S -> 'a : 'S

POP
:: 'a : 'S -> 'S

BRANCH: If 'a is true then push 'b else push 'c
:: 'a : 'b : 'c : 'S -> ('b or 'c) : 'S

GREATER_THAN:
:: 'a : 'b : 'S -> 'a > 'b : 'S

GREATER_THAN_OR_EQUAL:
:: 'a : 'b : 'S -> 'a >= 'b : 'S

EQUAL:
:: 'a : 'b : 'S -> 'a == 'b : 'S

NOT:
:: 'a:boolean : 'S -> 'not 'a : 'S


**MATH OPERATORS**

Math operators can be used via an integer or a token_amount.
You can divide a token amount by an integer, but not the other way around.

ADD
:: 'a : 'b : 'S -> 'a + 'b : 'S

SUB
:: 'a : 'b : 'S -> 'a - 'b : 'S

MUL
:: 'a : 'b : 'S -> 'a * 'b : 'S

DIV:
:: 'a : 'b : 'S -> 'a / 'b : 'S

PUSH 3
PUSH 1
DIV
// [0.333:float]


MOD
:: 'a : 'b : 'S -> 'a % 'b : 'S

**SPECIAL OPERATORS**

{special_operators_doc}

---

Your task is to convert a user's request into commands on the Dowse stack.

The stack is shown as the variable name, followed by the type, separated by a colon.
Make sure when you add an element, it goes to the left on the stack.

For example:
    # Stack: [0.001_wei:token_amount, cbBTC_address:address, eth_address:address]
---

{examples}

---

Since Dowse is a type safe language, you need to make sure the order of arguments on the stack
are the same as the function signature.  So a function like:

MY_COMMAND
:: ':integer : ':float : 'S -> ':float : 'S

requires that the top of the stack is an integer and the second element is a float.
If this is not the case, you need to swap the elements so they are in the correct order.
"""



class StackEmpty(Exception):
    pass


class CommandError(Exception):
    pass


class StackError(Exception):
    pass


class StackTypeError(StackError):
    pass


class StackValueError(StackError):
    pass



import logging
from typing import Awaitable, Callable, Generic, Self, Type, TypeVar

from emp_agents import AgentBase
from emp_agents.models import Message, Provider
from emp_agents.providers import OpenAIModelType, OpenAIProvider
from emp_agents.utils import count_tokens
from pydantic import BaseModel, Field, PrivateAttr

from dowse.models.message import AgentMessage

from ..loaders import Loaders
from .executor import Executor

logger = logging.getLogger("dowse")

T = TypeVar("T", bound=BaseModel)
OutputType = TypeVar("OutputType", bound=BaseModel | str | None)


class AgentExecutor(
    Executor[T, OutputType],
    Loaders,
    Generic[T, OutputType],
):
    provider: Provider = Field(
        default_factory=lambda: OpenAIProvider(default_model=OpenAIModelType.gpt4o)
    )
    tools: list[Callable] = Field(default_factory=list)
    examples: list[list[Message]] = Field(default_factory=list)

    _agent: AgentBase | None = PrivateAttr(default=None)

    def set_provider(self, provider: Provider) -> Self:
        self.provider = provider
        return self

    async def load_tools(self, input_: T) -> list[Callable]:
        """This can be overriden by subclasses to provide tools dynamically"""
        return self.tools

    async def after_execution(self, input_: T, output: OutputType) -> None:
        """A function that is called after the execution completes.
        Can be overridden by subclasses to add post-execution behavior.

        Args:
            input_: The original input that was executed
            output: The output from the execution
        """
        for effect in self.effects:
            maybe_coro = effect.execute(input_, output)
            if isinstance(maybe_coro, Awaitable):
                await maybe_coro

    async def execute(
        self,
        input_: T,
        persist: bool = False,
    ) -> AgentMessage[OutputType]:
        response_format: Type[OutputType] = self._extract_response_format()

        if response_format is None:
            return AgentMessage[response_format](  # type: ignore[valid-type]
                content="",
                error_message=None,
            )

        if not self.prompt:
            raise ValueError("Prompt is required")

        if (not persist) or (self._agent is None):
            self._agent = AgentBase(
                prompt=self.prompt,
                provider=self.provider,
                tools=await self.load_tools(input_),
                sync_tools=True,
            )

        for example_lst in self.examples:
            self._agent.add_messages(example_lst)

        try:
            response: OutputType = await self._agent.answer(
                input_.model_dump_json(),
                response_format=response_format,  # type: ignore[arg-type]
            )  # type: ignore[assignment]
        except Exception as e:
            return AgentMessage[OutputType](
                content=None,
                error_message=str(e),
            )
        await self.after_execution(input_, response)

        return AgentMessage[OutputType](
            content=response,
            error_message=None,
        )

    async def ask(self, question: str) -> str:
        """Ask the current agent a question about the execution"""
        if self._agent is None:
            raise ValueError("Agent is not initialized")
        return await self._agent.answer(question)

    def reset_agent(self):
        """Reset the agent to its initial state"""
        self._agent.reset()

    def _extract_response_format(self) -> type[OutputType]:
        if type(self).__base__ is AgentExecutor:
            return type(self).__pydantic_generic_metadata__["args"][1]
        return type(self).__bases__[0].__pydantic_generic_metadata__["args"][1]  # type: ignore[attr-defined]

    def __repr__(self):
        return f"{self.__class__.__name__}"

    __str__ = __repr__

    def token_count(self) -> int:
        if self.prompt is None:
            return 0
        return count_tokens(self.prompt)



from typing import TYPE_CHECKING, Awaitable, Generic, Self, TypeVar

from pydantic import BaseModel

from dowse.models.message import AgentMessage

from ..effects import Effect
from .base import ExecutorT

if TYPE_CHECKING:
    from .chain import ChainExecutor

T = TypeVar("T", bound=BaseModel | str | None)
OutputType = TypeVar("OutputType", bound=BaseModel | str | None)
U = TypeVar("U", bound=BaseModel | str | None)


class Executor(ExecutorT[T, OutputType], BaseModel, Generic[T, OutputType]):
    async def execute(
        self,
        input_: T,
        persist: bool = False,
    ) -> AgentMessage[OutputType]:
        # NOTE: pydantic does not allow ABCs as parameter type args so we need to raise an error
        raise NotImplementedError("Subclasses must implement this method")

    async def after_execution(self, input_: T, output: OutputType) -> None:
        """A function that is called after the execution completes.
        Can also be overridden by subclasses to add post-execution behavior.

        Args:
            input_: The original input that was executed
            output: The output from the execution
        """
        for effect in self.effects:
            maybe_coro = effect.execute(input_, output)
            if isinstance(maybe_coro, Awaitable):
                await maybe_coro

    def __rshift__(
        self,
        other: "Executor[OutputType, U]",
    ) -> "ChainExecutor[T, OutputType, U]":
        from .chain import ChainExecutor

        return ChainExecutor(left_executor=self, right_executor=other)

    def add_effect(self, effect: Effect[T, OutputType]) -> Self:
        self.effects.append(effect)
        return self

    def add_effects(self, effects: list[Effect[T, OutputType]]) -> Self:
        self.effects.extend(effects)
        return self



EXAMPLE_PROGRAM = """
The command "Swap 0.001 ETH for $AERO and send half to @user2 and half to @user3"
can be executed with the following code:

```code
PUSH "AERO"
GET_TOKEN_ADDRESS
PUSH 0.001
CONVERT_ETH_TO_WEI
// ['0.001_wei:token_amount 'aero_address:address]
EXCHANGE_FUNDS
// ['0.001_aero:token_amount]
PUSH 2
SWAP
DIV
ASSIGN aero_balance
// ['0.0005_aero:token_amount]
PUSH @user2
SWAP
// ['0.0005_aero:token_amount '@user2]
TRANSFER_FUNDS
PUSH @user3
PUSH &aero_balance
TRANSFER_FUNDS
// ['0.0005_aero:token_amount '0.0005_aero:token_amount]
```

---

Example Functions:

```code
PUSH 300
PUSH 200
PUSH 100
PUSH 500
// ['500:integer '100:integer '200:integer '300:integer]
DIV
// ['5:integer '200:integer '300:integer]
PUSH 3
GREATER_THAN
// ['false:boolean '200:integer '300:integer]
BRANCH
// ['300:integer]
```

Ensure the stack is empty at the end, with no unused values.
Return only the code without any additional text.

Show the stack after each operation.
"""



from ..types import Address, Boolean, Float, String, TokenAmount, User
from .stack_op import StackOp


class MakeUser(StackOp):
    """Convert a username to a User"""

    def operation(self, username: String | User) -> User:
        if isinstance(username, User):  # type: ignore[misc]
            return username
        return User(username.value)


class UserToAddress(StackOp):
    """Convert a User to an Address"""

    def operation(self, user: User) -> Address:
        return Address(user.value)


class MakeTokenAmount(StackOp):
    """Convert a Float and Address to a TokenAmount"""

    def operation(self, amount: Float, address: Address) -> TokenAmount:
        return TokenAmount((amount, address))


class GetTokenAddress(StackOp):
    """Get the Address of a Token"""

    def operation(self, symbol: String) -> Address:
        return Address(symbol.value)


class ConvertEthToWei(StackOp):
    """Convert a Float to an ETH TokenAmount"""

    def operation(self, amount: Float) -> TokenAmount:
        return TokenAmount(
            (amount, Address("0x4200000000000000000000000000000000000006"))
        )


class ConvertToTokenAmount(StackOp):
    """Convert an Amount and Address to a TokenAmount"""

    def operation(self, amount: Float, address: Address) -> TokenAmount:
        return TokenAmount((amount, address))


class TransferFunds(StackOp):
    """Transfer funds from the current address to the given address, pushing the new TokenAmount onto the stack"""

    def operation(self, amount: TokenAmount, to_address: Address | User) -> TokenAmount:
        return TokenAmount((amount.value[0], to_address))


class MaybeTransferFunds(StackOp):
    """
    Does the transfer if the third element in the stack is true, pushing the new TokenAmount onto the stack.
    Otherwise pushes the original TokenAmount back onto the stack."""

    def operation(
        self,
        amount: TokenAmount,
        to_address: Address | User,
        do_transfer: Boolean,
    ) -> TokenAmount:
        if do_transfer.value:
            return TokenAmount((amount.value[0], to_address))
        return amount


class ExchangeFunds(StackOp):
    """Convert a TokenAmount to another Address via a swap, buy, sell, exchange, etc."""

    def operation(self, amount: TokenAmount, to_address: Address) -> TokenAmount:
        return TokenAmount((amount.value[0], to_address))


class GetPercentage(StackOp):
    """
    Get a percentage of a TokenAmount.  First arg must be a TokenAmount, second arg must be a Float between 0 and 1.
    Splits the Amount into two TokenAmounts, one with the percentage and one with the remainder.
    """

    def operation(
        self, amount: TokenAmount, percentage: Float
    ) -> tuple[TokenAmount, TokenAmount]:
        return (
            TokenAmount((amount.value[0] * percentage.value, amount.value[1])),  # type: ignore[arg-type, operator]
            TokenAmount(
                (amount.value[0] * (1 - percentage.value), amount.value[1])  # type: ignore[arg-type, operator]
            ),
        )


SPECIAL_OPERATORS: list = [
    MakeUser,
    UserToAddress,
    MakeTokenAmount,
    GetTokenAddress,
    ConvertEthToWei,
    ConvertToTokenAmount,
    TransferFunds,
    MaybeTransferFunds,
    ExchangeFunds,
    GetPercentage,
]



from typing import Callable

from pydantic import BaseModel, Field

from dowse.interfaces import AgentExecutor
from dowse.models import Tweet
from dowse.tools import (
    convert_dollar_amount_to_eth,
    convert_token_amount_to_wei,
    get_token_address_tool,
)
from dowse.tools.wei import convert_decimal_eth_to_wei


class FormattedCommand(BaseModel):
    content: str
    caller: str = Field(description="The Twitter handle of the caller")


class ProcessTokens(AgentExecutor[Tweet, FormattedCommand]):
    tools: list[Callable] = [
        get_token_address_tool,
        convert_dollar_amount_to_eth,
        convert_token_amount_to_wei,
        convert_decimal_eth_to_wei,
    ]


process_tokens = ProcessTokens()



import json

from emp_agents.models import AssistantMessage, ToolCall, ToolMessage, UserMessage

EXAMPLES = [
    [
        UserMessage(
            content=json.dumps(
                {
                    "caller": "@user",
                    "content": (
                        "#simmi Whats up?  I want you to swap $123.13 for AERO"
                        " and then send half of it to @user2.  Can you do that for me now?"
                    ),
                }
            )
        ),
        AssistantMessage(
            content=(
                'Ok, I can convert simplify this command to "1. swap $123.13 for $AERO\n'
                '2. transfer 50% of the output to @user2" and then use my tools to further refine'
            ),
            tool_calls=[
                ToolCall(
                    id="57d74922-3636-49b1-8109-fe25ca66ca19",
                    type="function",
                    function=ToolCall.Function(
                        name="get_token_address_tool",
                        arguments='{"symbol_or_token_address": "AERO"}',
                    ),
                ),
                ToolCall(
                    id="0b73576d-413a-4fb8-8fe4-99da789fb0e1",
                    type="function",
                    function=ToolCall.Function(
                        name="convert_dollar_amount_to_eth",
                        arguments='{"amount": "$123.13"}',
                    ),
                ),
            ],
        ),
        ToolMessage(
            tool_call_id="57d74922-3636-49b1-8109-fe25ca66ca19",
            content="0x940181a94A35A4569E4529A3CDfB74e38FD98631",
        ),
        ToolMessage(
            tool_call_id="0b73576d-413a-4fb8-8fe4-99da789fb0e1",
            content="46200000000000000 ETH",
        ),
        AssistantMessage(
            content=json.dumps(
                {
                    "caller": "@user",
                    "user_request": (
                        "1. swap 46200000000000000 ETH (0x4200000000000000000000000000000000000006)"
                        " for $AERO (0x940181a94A35A4569E4529A3CDfB74e38FD98631)\n"
                        "2. send 50% of the output to @user2"
                    ),
                }
            ),
        ),
        UserMessage(
            content="Ok Great, lets try this for a more complicated case",
        ),
        AssistantMessage(content="I understand."),
    ],
]



from importlib.util import module_from_spec, spec_from_file_location
from typing import Any, Callable

from pydantic import Field

from .base import Loader


class ToolsLoader(Loader):
    tools: list[Callable] = Field(default_factory=list)

    def model_post_init(self, __context: Any) -> None:
        super().model_post_init(__context)

        tools_path = self._path / "tools.py"
        if tools_path.exists():

            spec = spec_from_file_location("tools", tools_path)
            if spec and spec.loader:
                module = module_from_spec(spec)
                spec.loader.exec_module(module)
                if hasattr(module, "TOOLS"):
                    self.tools = module.TOOLS



from typing import Generic, TypeVar

T = TypeVar("T")


class DowseType(Generic[T]):
    name: str
    description: str

    value: T

    def __init__(self, value: T):
        self.value = value

    def __repr__(self) -> str:
        return f"'{self.value}:{self.name}"

    __str__ = __repr__


class String(DowseType[str]):
    name = "string"
    description = "A string"


class User(DowseType, str):
    name = "user"
    description = (
        "A user.  Denoted by @username with no quotes, spaces or special characters."
    )


class Integer(DowseType[int]):
    name = "integer"
    description = "An integer"


class Float(DowseType[float]):
    name = "float"
    description = "A float"


class Address(DowseType[str]):
    name = "address"
    description = "An Ethereum address"


class TokenAmount(DowseType[tuple[Float, Address | User]]):
    name = "token_amount"
    description = "A token amount"


TokenAmount.__repr__ = lambda self: f"'{self.value[0].value}:{self.title}"  # type: ignore


DEFAULT_TYPES = [
    Address,
    TokenAmount,
    String,
    Integer,
    Float,
    Boolean,
]


def foo(s: String):
    pass


foo(String("3"))



from eth_rpc.networks import Base
from eth_typeshed.erc20 import ERC20

from dowse.logger import logger


async def convert_token_amount_to_wei(amount: str, token_address: str) -> str:
    """
    convert a decimal token amount to wei
    """
    token = ERC20[Base](address=token_address)
    decimals = await token.decimals().get()
    return str(int(float(amount) * 10**decimals))


def convert_decimal_eth_to_wei(amount: str) -> str:
    """
    convert a decimal ETH amount to a string ETH amount.

    For example, an amount like 0.00123 ETH should be converted to 1230000000000000
    """
    result = str(int(float(amount) * 1e18)) + " ETH"
    logger.debug("Tool Call: convert_decimal_eth_to_wei (%s) -> '%s'", amount, result)
    return result



from typing import Any

from pydantic import BaseModel, Field

from .exceptions import StackEmpty, StackValueError
from .operators.stack_op import StackOp
from .types import Wrapper


class Stack(BaseModel):
    stack: list[Wrapper[Any]] = Field(default_factory=list)
    operators: dict[str, "StackOp"] = Field(default_factory=dict)

    variables: dict[str, Wrapper[Any]] = Field(default_factory=dict)

    def roll(self, n: int) -> None:
        nth_element = self.stack[n]
        self.stack = self.stack[:n] + self.stack[n + 1 :] + [nth_element]

    def assign(self, variable_name: str, value: Wrapper[Any]) -> None:
        self.variables[variable_name] = value

    def push(self, *values: Wrapper[Any]) -> None:
        self.stack.extend(values)

    def pop(self) -> Wrapper[Any]:
        if self.is_empty():
            raise StackEmpty("tried to pop from an empty stack")
        return self.stack.pop()

    def peek(self) -> Wrapper[Any]:
        if self.is_empty():
            raise StackEmpty("tried to peek at an empty stack")
        return self.stack[-1]

    def is_empty(self) -> bool:
        return len(self.stack) == 0

    def execute_line(self, line: str) -> None:
        line = line.split("//")[0].strip()
        if not line:
            return
        line, *args = line.split()
        self.call(line, *args)

    def call(self, command: str, *args: Any) -> None:
        if command == "ROLL":
            stack_arg = self.pop()
            self.roll(stack_arg.value)
            return
        if command == "PUSH" and args and args[0].startswith("&"):
            variable_name = args[0][1:]
            if variable_name not in self.variables:
                raise StackValueError(f"Variable {variable_name} not found")
            self.push(self.variables[variable_name])
            return
        elif command == "ASSIGN":
            variable_name = args[0]
            stack_arg = self.pop()
            self.assign(variable_name, stack_arg)
            self.push(stack_arg)
            return

        function = self.operators[command]

        if function.immediate_instruction:
            result = function.operation(*args)
        else:
            stack_args = [self.pop() for _ in range(function.arg_count)]
            try:
                function.type_check(*stack_args)
                result = function.operation(*stack_args)
            except Exception as e:
                for arg in stack_args[::-1]:
                    self.push(arg)
                raise e
        if result is not None:
            if isinstance(result, tuple):
                for result_arg in result[::-1]:
                    self.push(result_arg)
            else:
                self.push(result)

    def register_operators(self, operators: list["StackOp"]) -> None:
        self.operators = {operator.name(): operator for operator in operators}

    def __repr__(self) -> str:
        return f"{self.stack[::-1]}"

    __str__ = __repr__



from .base import Loader
from .examples import ExampleLoader
from .prompts import PromptLoader
from .tools import ToolsLoader


class Loaders(ExampleLoader, PromptLoader, ToolsLoader):
    pass


__all__ = ["Loader", "ExampleLoader", "PromptLoader", "ToolsLoader", "Loaders"]



import logging
from typing import Generic, TypeVar

from pydantic import BaseModel, Field

from dowse.exceptions import PreprocessorError
from dowse.models.message import AgentMessage

from ..loaders import Loaders
from .agent import Processor

T = TypeVar("T")
U = TypeVar("U", bound=BaseModel)

logger = logging.getLogger("dowse")


class PreProcess(Loaders, Generic[T, U]):
    processors: list[Processor] = Field(default_factory=list)

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

    async def run_processors(self, input_: T) -> AgentMessage[U]:
        processed_data = input_
        if not self.processors:
            return AgentMessage(content=processed_data, error_message=None)  # type: ignore[arg-type]

        for processor in self.processors:
            logger.debug("Running preprocessor: %s", processed_data)
            processed_data: AgentMessage = await processor.process(processed_data)  # type: ignore[no-redef]
            logger.debug("Preprocessor result: %s", processed_data)
            if processed_data.error_message is not None:  # type: ignore[attr-defined]
                raise PreprocessorError(processed_data.error_message)  # type: ignore[attr-defined]
        return processed_data  # type: ignore



from typing import Any, Iterator

from pydantic import BaseModel, ConfigDict, computed_field
from pydantic._internal._model_construction import ModelMetaclass

from dowse.dsl.example import EXAMPLE_PROGRAM
from dowse.dsl.exceptions import StackEmpty, StackError, StackTypeError, StackValueError
from dowse.dsl.operators import BASE_OPERATORS, StackOp
from dowse.dsl.stack import Stack
from dowse.dsl.syntax import make_syntax_doc
from dowse.dsl.types import DEFAULT_TYPES
from dowse.interfaces import AgentExecutor


class UserRequest(BaseModel):
    content: str
    username: str


class Command(BaseModel):
    code_block: list[str]

    def print(self) -> None:
        idx = 0
        for line in self.code_block:
            if line.split("//")[0].strip() == "":
                continue
            print(f"{idx}: {line}")
            idx += 1

    def lines(self) -> Iterator[str]:
        for line in self.code_block:
            if line.startswith("//"):
                continue
            yield line


class DowseExecutor(
    AgentExecutor[
        UserRequest,  # the input type of the first processor
        Command,  # the output type of the last processor
    ]
):
    prompt: str = ""
    types: list[ModelMetaclass] = DEFAULT_TYPES
    base_operators: list[type[StackOp]] = BASE_OPERATORS
    special_operators: list[type[StackOp]] = []
    example_programs: list[str] = [EXAMPLE_PROGRAM]

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @computed_field  # type: ignore[misc]
    @property
    def operators(self) -> list[StackOp]:
        return [op() for op in self.base_operators + self.special_operators]

    @computed_field  # type: ignore[misc]
    @property
    def operator_map(self) -> dict[str, StackOp]:
        return {op.name(): op for op in self.operators}

    def model_post_init(self, __context: Any):
        self.prompt = make_syntax_doc(
            self.types,
            "\n---\n".join(self.example_programs),
            special_operators=self.special_operators,
        )
        super().model_post_init(__context)

    def simulate(self, command: Command, verbose: bool = False) -> Stack:
        stack = Stack()
        stack.register_operators(self.operators)

        for idx, line in enumerate(command.lines()):
            try:
                stack.execute_line(line)
            except StackTypeError as e:
                raise StackError(idx, line, e, stack)
            except StackValueError as e:
                raise StackError(idx, line, e, stack)
            except StackEmpty as e:
                raise StackError(idx, line, e, stack)
            if verbose:
                print(line)
                print("Stack: ", stack)
        return stack

    async def solve(
        self,
        query: UserRequest,
        max_errors: int = 10,
        verbose: bool = False,
    ) -> Command:
        error_count = 0
        response = await self.execute(
            query,
            persist=True,
        )
        while error_count < max_errors:
            if response.content is None:
                raise Exception(f"Execution Error: {response.error_message}")

            code = response.content
            if verbose:
                print("CODE BLOCK:")
                code.print()
            try:
                self.simulate(code, verbose=False)
                return code

            except StackError as e:
                (idx, line, error, stack) = e.args
                error_str = f"Error executing line {idx}:{line}: {error}.  Stack currently: {stack}"

                if verbose:
                    print(f"\n\nERROR:{error_str}\n\n")

                response = await self.execute(
                    UserRequest(
                        content=f"{error_str}\nPlease fix program and try again",
                        username=query.username,
                    ),
                    persist=True,
                )
        raise Exception("Max errors reached")



from typing import Any

from .base import Loader


class PromptLoader(Loader):
    prompt: str | None = None

    def model_post_init(self, __context: Any) -> None:
        super().model_post_init(__context)

        prompt_path = self._path / "PROMPT.txt"
        if prompt_path.exists():
            self.prompt = prompt_path.read_text()



from emp_agents.providers import OpenAIModelType, OpenAIProvider
from pydantic import BaseModel

from dowse.interfaces import AgentExecutor

from ...tools import CommandRequest, Tools
from ..preprocessor import FormattedCommand, process_tokens


class CommandsList(BaseModel):
    commands: list[CommandRequest]


BasicTwitterCommands = process_tokens >> AgentExecutor[FormattedCommand, CommandsList](
    provider=OpenAIProvider(
        default_model=OpenAIModelType.gpt4o,
    ),
    tools=[
        Tools.get_amount_out_tool,
        Tools.get_percentage,
    ],
)



from abc import ABC, abstractmethod
from typing import Generic, TypeVar

from pydantic import BaseModel

from dowse.models import AgentMessage

T = TypeVar("T")
U = TypeVar("U")


class Processor(ABC, BaseModel, Generic[T, U]):
    @abstractmethod
    async def process(self, command: T) -> AgentMessage[U]: ...



from typing import Generic, TypeVar

from pydantic import BaseModel, ConfigDict

from dowse.models.message import AgentMessage

from .executor import Executor, ExecutorT

T = TypeVar("T")
InputType = TypeVar("InputType", bound=BaseModel | str | None)
IntermediateType = TypeVar("IntermediateType", bound=BaseModel | str | None)
OutputType = TypeVar("OutputType", bound=BaseModel | str | None)


class ChainExecutor(
    Executor[InputType, OutputType],
    BaseModel,
    Generic[InputType, IntermediateType, OutputType],
):
    left_executor: ExecutorT[InputType, IntermediateType]
    right_executor: ExecutorT[IntermediateType, OutputType]

    model_config = ConfigDict(
        arbitrary_types_allowed=True,  # Needed to allow non-BaseModel types
    )

    async def execute(
        self,
        input_: InputType,
        persist: bool = False,
    ) -> AgentMessage[OutputType]:
        print(self.left_executor, type(self.left_executor))
        print(self.right_executor, type(self.right_executor))
        intermediate = await self.left_executor.execute(input_, persist=persist)

        if intermediate.error_message:
            return AgentMessage[OutputType](
                content=None,
                error_message=intermediate.error_message,
            )

        assert intermediate.content is not None
        return await self.right_executor.execute(intermediate.content, persist=persist)



from importlib.util import module_from_spec, spec_from_file_location
from typing import Any

from emp_agents.models import Message
from pydantic import Field

from .base import Loader


class ExampleLoader(Loader):
    examples: list[list[Message]] = Field(default_factory=list)

    def model_post_init(self, __context: Any) -> None:
        super().model_post_init(__context)

        examples_path = self._path / "examples.py"
        if examples_path.exists():

            spec = spec_from_file_location("examples", examples_path)
            if spec and spec.loader:
                module = module_from_spec(spec)
                spec.loader.exec_module(module)
                if hasattr(module, "EXAMPLES"):
                    self.examples = module.EXAMPLES



from abc import ABC, abstractmethod
from typing import Generic, TypeVar

from pydantic import BaseModel, Field

from dowse.models.message import AgentMessage

from ..effects import Effect

T = TypeVar("T", bound=BaseModel | str | None)
OutputType = TypeVar("OutputType", bound=BaseModel | str | None)


class ExecutorT(ABC, Generic[T, OutputType]):
    effects: list[Effect[T, OutputType]] = Field(default_factory=list)

    @abstractmethod
    async def execute(
        self,
        input_: T,
        persist: bool = False,
    ) -> AgentMessage[OutputType]:
        """
        Map the input to the output type
        """

    @abstractmethod
    async def after_execution(self, input_: T, output: OutputType) -> None:
        """A function that is called after the execution completes.
        Can also be overridden by subclasses to add post-execution behavior.

        Args:
            input_: The original input that was executed
            output: The output from the execution
        """



from typing import Annotated, Generic, TypeVar, get_args

from pydantic import BaseModel, Field

T = TypeVar("T")


class Wrapper(BaseModel, Generic[T]):
    value: T

    def __init__(self, value: T):
        super().__init__(value=value)

    def __repr__(self) -> str:
        return f"'{self.value}:{self.title}"

    __str__ = __repr__

    @classmethod
    def _annotation(cls):
        return get_args(cls.__pydantic_generic_metadata__["args"][0])[1]

    @property
    def title(self):
        return self._annotation().title

    @property
    def description(self):
        return self._annotation().description

    @classmethod
    def name(cls):
        try:
            return cls._annotation().title
        except Exception:
            return str(cls.__pydantic_generic_metadata__["args"][0]).split(".")[-1]


String = Wrapper[Annotated[str, Field(title="string", description="A string")]]
User = Wrapper[
    Annotated[
        str,
        Field(
            title="user",
            description="A user.  Denoted by @username with no quotes, spaces or special characters.",
        ),
    ]
]
Integer = Wrapper[Annotated[int, Field(title="integer", description="An integer")]]
Float = Wrapper[Annotated[float, Field(title="float", description="A float")]]
Boolean = Wrapper[Annotated[bool, Field(title="boolean", description="A boolean")]]
Address = Wrapper[
    Annotated[str, Field(title="address", description="An Ethereum address")]
]
TokenAmount = Wrapper[
    Annotated[
        tuple[Float, Address | User],
        Field(title="token_amount", description="A token amount"),
    ]
]
TokenAmount.__repr__ = lambda self: f"'{self.value[0].value}:{self.title}"  # type: ignore


DEFAULT_TYPES = [
    Address,
    TokenAmount,
    String,
    Integer,
    Float,
    Boolean,
]



import inspect
from abc import ABC
from textwrap import dedent
from types import UnionType
from typing import Any, ClassVar, get_args, get_origin

from pydantic import BaseModel

from dowse.dsl.exceptions import StackTypeError

from ..types import Wrapper


class StackOp(ABC, BaseModel):
    immediate_instruction: ClassVar[bool] = (
        False  # Operations that can take args like 'PUSH'
    )

    def type_check(self, *args: Any) -> None:
        signature = self.signature()
        for index, (arg, arg_type) in enumerate(zip(args, self.arg_types)):
            if not isinstance(arg, arg_type):

                raise StackTypeError(
                    f"Argument {index} to {signature} is of type {arg.title}"
                )

    def operation(self, *args: Any) -> Wrapper[Any] | tuple[Wrapper[Any], ...] | None:
        return None

    @property
    def arg_count(self) -> int:
        return len(inspect.signature(self.operation).parameters)

    @property
    def arg_types(self):
        return [
            arg.annotation
            for arg in inspect.signature(self.operation).parameters.values()
        ]

    def signature(self):
        signature = f"{self.name()} "
        for arg in self.arg_types:
            if get_origin(arg) is UnionType:
                union = "("
                for type_ in get_args(arg):
                    union += f"{type_.name()} | "
                union = union.strip(" | ")
                signature += f"{union}) "
            else:
                signature += f"{arg.name()} "
        signature = signature.strip()
        return signature

    @classmethod
    def name(cls) -> str:
        name = cls.__name__
        return (
            "".join(["_" + i.lower() if i.isupper() else i for i in name])
            .upper()
            .lstrip("_")
        )

    def __repr__(self) -> str:
        return f"{self.name()}"

    __str__ = __repr__

    @classmethod
    def to_docstring(cls) -> str:
        func_signature = ":: "
        annotations = cls.operation.__annotations__
        func_name = cls.name()
        docstring = cls.__doc__
        for name, annotation in annotations.items():
            if name == "return":
                break

            if get_origin(annotation) is UnionType:
                union = "("
                for type_ in get_args(annotation):
                    union += f"{type_.name()} | "
                union = union.strip(" | ")
                func_signature += f"{union}) "
            else:
                func_signature += f"':{annotation._annotation().title} : "
        func_signature += "'S -> "
        return_type = annotations["return"]

        if return_type is None:
            func_signature += "'S"
        else:
            func_signature += f"'{return_type._annotation().title} : 'S"
        return (
            dedent(
                f"""
            {func_name}: {docstring}
            {func_signature}
        """
            ).strip()
            + "\n"
        )



# flake8: noqa: E501

import json

from emp_agents.models import AssistantMessage, ToolCall, ToolMessage, UserMessage

EXAMPLES = [
    [
        UserMessage(
            content="{'caller': '@jack', 'content': '1. swap 10000000000000000 ETH (0x4200000000000000000000000000000006) for $AERO (0x940181a94A35A4569E4529A3CDfB74e38FD98631)'}"
        ),
        AssistantMessage(
            content="Lets determine the amount of tokens that will be the output of this swap",
            tool_calls=[
                ToolCall(
                    id="c4649732-4cc9-47e4-bc48-88224566686b",
                    type="function",
                    function=ToolCall.Function(
                        name="get_amount_out_tool",
                        arguments='{"token_in_address": "0x4200000000000000000000000000000006", "token_out_address": "0x940181a94A35A4569E4529A3CDfB74e38FD98631", "amount": "10000000000000000"}',
                    ),
                )
            ],
        ),
        ToolMessage(
            tool_call_id="c4649732-4cc9-47e4-bc48-88224566686b",
            content="1394601928424631358765",
        ),
        AssistantMessage(
            content="""
            {
                "commands": [
                    {"command": "swap", "args": {"token_in_address": "0x4200000000000000000000000000000006", "token_out_address": "0x940181a94A35A4569E4529A3CDfB74e38FD98631", "amount_in": "3000000000000000", "amount_out": "1394601928424631358765", "recipient": "@jack"}},
                ],
                "error_message": None,
            }
            """,
        ),
    ],
    [
        UserMessage(
            content="{'caller': '@abc123', 'content': 'swap 3000000000000000 ETH (0x4200000000000000000000000000000000000006) for $AIXBT (0x4F9Fd6Be4a90f2620860d680c0d4d5Fb53d1A825) and send half of it to @myfriend'}"
        ),
        AssistantMessage(
            content="Lets determine the amount of tokens that will be the output of this swap",
            tool_calls=[
                ToolCall(
                    id="c4649732-4cc9-47e4-bc48-88224566686b",
                    type="function",
                    function=ToolCall.Function(
                        name="get_amount_out_tool",
                        arguments='{"token_in_address": "0x4200000000000000000000000000000000000006", "token_out_address": "0x4F9Fd6Be4a90f2620860d680c0d4d5Fb53d1A825", "amount": "3000000000000000"}',
                    ),
                )
            ],
        ),
        ToolMessage(
            tool_call_id="c4649732-4cc9-47e4-bc48-88224566686b",
            content="1394601928424631358765",
        ),
        AssistantMessage(
            content="Lets use the percentage tool to get half of the amount out",
            tool_calls=[
                ToolCall(
                    id="c4649732-4cc9-47e4-bc48-88224566686b",
                    type="function",
                    function=ToolCall.Function(
                        name="get_percentage",
                        arguments='{"amount": "1394601928424631358765", "percentage": "50"}',
                    ),
                )
            ],
        ),
        ToolMessage(
            tool_call_id="c4649732-4cc9-47e4-bc48-88224566686b",
            content="697300964212315679382",
        ),
        AssistantMessage(
            content="""
            {
                "commands": [
                    {"command": "swap", "args": {"token_in_address": "0x4200000000000000000000000000000006", "token_out_address": "0x4F9Fd6Be4a90f2620860d680c0d4d5Fb53d1A825", "amount_in": "3000000000000000", "amount_out": "1394601928424631358765", "recipient": "@user"}},
                    {"command": "transfer", "args": {"token_address": "0x4F9Fd6Be4a90f2620860d680c0d4d5Fb53d1A825", "amount": "697300964212315679382", "recipient": "@myfriend"}}
                ],
            }
            """,
        ),
    ],
    [
        UserMessage(content="ok now i'm going to send you a new user's commands"),
        AssistantMessage(
            content="ok, i'll wait",
        ),
    ],
]



from typing import Any

from dowse.interfaces import Executor
from dowse.models import AgentMessage


class NoOpExecutor(Executor[Any, None]):
    async def execute(
        self,
        input_: Any,
        persist: bool = False,
    ) -> AgentMessage[None]:
        return AgentMessage(content=None, error_message=None)



import inspect
from pathlib import Path
from typing import Any

from pydantic import BaseModel, PrivateAttr


class Loader(BaseModel):
    _path: Path = PrivateAttr()

    def model_post_init(self, __context: Any) -> None:
        stack = inspect.stack()

        parent_path = Path(__file__).parent.parent

        caller_frame = None
        for frame_info in stack:
            filename = frame_info.filename
            if filename.startswith("<frozen"):
                continue
            if Path(filename).is_relative_to(parent_path):
                continue
            if "pydantic" not in filename:
                caller_frame = frame_info
                break

        assert caller_frame is not None
        self._path = Path(caller_frame.filename).parent
